# Copyright (c) 2008-2016 Szczepan Faber, Serhiy Oplakanets, Herr Kaste
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import annotations
import inspect
import operator
import types
import functools
from collections import deque
from contextlib import contextmanager
from dataclasses import dataclass
from typing import Any, AsyncIterator, Callable, Iterable, Iterator, cast

from . import invocation, sameish, signature, utils
from . import verification as verificationModule
from .mock_registry import mock_registry
from .patching import Patch, patcher


__all__ = ['mock']

__tracebackhide__ = operator.methodcaller(
    "errisinstance",
    invocation.InvocationError
)
SUPPORTS_MARKCOROUTINEFUNCTION = hasattr(inspect, "markcoroutinefunction")

_CONFIG_ASYNC_PREFIX = "async "
_ASYNC_BY_PROTOCOL_METHODS = {"__aenter__", "__aexit__", "__anext__"}


class _Dummy:
    # We spell out `__call__` here for convenience. All other magic methods
    # must be configured before use, but we want `mock`s to be callable by
    # default.
    def __call__(self, *args, **kwargs):
        return self.__getattr__('__call__')(*args, **kwargs)  # type: ignore[attr-defined]  # noqa: E501


def remembered_invocation_builder(
    mock: Mock,
    method_name: str,
    discard_first_arg: bool,
    *args,
    **kwargs
):
    invoc = invocation.RememberedInvocation(
        mock, method_name, discard_first_arg=discard_first_arg
    )
    return invoc(*args, **kwargs)


ANSWER_SELECTOR_METHODS = {
    'thenReturn',
    'thenRaise',
    'thenAnswer',
    'thenCallOriginalImplementation',
}


@dataclass(frozen=True)
class Chain:
    theMock: Mock
    options: dict[str, Any]
    segments: tuple[Segment, ...] = ()

    def __add__(self, segment: Segment) -> "Chain":
        return Chain(self.theMock, self.options, self.segments + (segment,))

    def new_segment(self, name: str):
        return _chain_segment(self, name)

    def rollback(self) -> None:
        for segment in reversed(self.segments):
            segment.invoc.forget_self()


@dataclass(frozen=True)
class Segment:
    name: str
    invoc: invocation.StubbedInvocation
    answer_selector: invocation.AnswerSelector


def _chain_segment(chain: Chain, name: str):
    class SegmentFacade:
        def __call__(self, *args, **kwargs):
            if chain.segments and name in ANSWER_SELECTOR_METHODS:
                getattr(chain.segments[-1].answer_selector, name)(
                    *args,
                    **kwargs,
                )
                return _wait_for_attr(chain)

            segment = _materialize_method_segment(chain, name, args, kwargs)
            next_chain = chain + segment
            return _wait_for_attr(next_chain)

        def __getattr__(self, attr_name):
            try:
                segment = _materialize_property_segment(chain, name)
            except invocation.InvocationError:
                chain.rollback()
                raise

            next_chain = chain + segment
            return next_chain.new_segment(attr_name)

        def __enter__(self):
            chain.rollback()
            raise AttributeError('__enter__')

        def __exit__(self, *exc_info):
            chain.rollback()
            raise AttributeError('__exit__')

    return SegmentFacade()


def _wait_for_attr(chain: Chain):
    class WaitForAttr:
        def __getattr__(self, attr_name):
            return chain.new_segment(attr_name)

        __enter__ = chain.segments[-1].answer_selector.__enter__
        __exit__ = chain.segments[-1].answer_selector.__exit__

    return WaitForAttr()


def _materialize_method_segment(
    chain: Chain,
    name: str,
    args: tuple[Any, ...],
    kwargs: dict[str, Any],
) -> Segment:
    if not chain.segments:
        # Pre-checks here are strict-independent and validate fluent-DSL
        # *shape* (e.g. call-vs-attribute intent) before we patch anything.
        # Strict/spec validation remains in StubbedInvocation.__call__
        # (`ensure_mocked_object_has_method` + signature checks).
        _ensure_target_is_callable(chain.theMock, name)

        invoc = invocation.StubbedInvocation(chain.theMock, name, **chain.options)
        answer_selector = invoc(*args, **kwargs)
        return Segment(name, invoc, answer_selector)

    chain_mock, parent_invocation, verification = _transition_to_child_chain(
        chain.segments
    )
    invoc = invocation.StubbedInvocation(
        chain_mock,
        name,
        verification=verification,
        parent_invocation=parent_invocation,
    )
    answer_selector = invoc(*args, **kwargs)
    return Segment(name, invoc, answer_selector)


def _materialize_property_segment(
    chain: Chain,
    name: str,
) -> Segment:
    if not chain.segments:
        # Same rationale as method materialization above: these pre-checks are
        # strict-independent API-shape guards for property-style stubbing.
        # Strict/spec checks for property stubs live in
        # StubbedPropertyAccess.__call__ (`ensure_mocked_object_has_attribute`).
        _ensure_target_is_not_callable(chain.theMock, name)

        if not inspect.isclass(chain.theMock.mocked_obj):
            raise invocation.InvocationError(
                "Cannot stub property '%s' on an instance. "
                "Use class-level stubbing instead: "
                "when(%s).%s.thenReturn(...)."
                % (
                    name,
                    type(chain.theMock.mocked_obj).__name__,
                    name,
                )
            )

        invoc = invocation.StubbedPropertyAccess(
            chain.theMock,
            name,
            **chain.options,
        )
        answer_selector = invoc()
        return Segment(name, invoc, answer_selector)

    chain_mock, parent_invocation, verification = _transition_to_child_chain(
        chain.segments
    )
    invoc = invocation.StubbedPropertyAccess(
        chain_mock,
        name,
        verification=verification,
        parent_invocation=parent_invocation,
    )
    answer_selector = invoc()
    return Segment(name, invoc, answer_selector)


def _transition_to_child_chain(
    segments: tuple[Segment, ...],
) -> tuple[
    Mock,
    invocation.StubbedInvocation,
    verificationModule.VerificationMode | None,
]:
    parent_invocation = segments[-1].invoc
    continuation = parent_invocation.transition_to_chain()
    verification = parent_invocation.pop_verification()
    return continuation.chain_mock, continuation.invocation, verification


def _ensure_target_is_callable(theMock: Mock, method_name: str) -> None:
    target, was_in_spec = theMock._get_original_method_before_stub(method_name)

    # Missing attributes can still be added in loose mode.
    if not was_in_spec and target is None:
        return

    if _should_continue_with_stubbed_invocation(
        target,
        allow_classes=True,
        spec=theMock.spec,
    ):
        return

    raise invocation.InvocationError("'%s' is not callable." % method_name)


def _ensure_target_is_not_callable(theMock: Mock, method_name: str) -> None:
    spec = theMock.spec
    if spec is None:
        return

    try:
        value = inspect.getattr_static(spec, method_name)
    except AttributeError:
        if inspect.isclass(spec):
            try:
                value = getattr(spec, method_name)
            except AttributeError:
                return
        else:
            return

    if _should_continue_with_stubbed_invocation(value, spec=spec):
        raise invocation.InvocationError(
            f"expected an invocation of '{method_name}'"
        )


def _should_continue_with_stubbed_invocation(
    value: object,
    allow_classes: bool = False,
    spec: object | None = None,
) -> bool:
    if (
        inspect.isfunction(value)
        or inspect.ismethod(value)
        or inspect.isbuiltin(value)
        or isinstance(value, staticmethod)
        or isinstance(value, classmethod)
        or isinstance(value, functools.partialmethod)
        or isinstance(value, types.MethodDescriptorType)
        or isinstance(value, types.WrapperDescriptorType)
        or isinstance(value, types.ClassMethodDescriptorType)
    ):
        return True

    # For class specs, callable descriptors (objects implementing both
    # `__call__` and `__get__`) are generally meant to be stubbed through
    # the property path. For non-class specs (e.g. module attributes such as
    # `numpy.vstack`), `__get__` should not disqualify callable targets.
    treat_callable_descriptors_as_non_callable = inspect.isclass(spec)

    # Generic callable fallback, with optional handling for callable
    # descriptor-like objects (`__call__` + `__get__`).
    return (
        callable(value)
        and (allow_classes or not inspect.isclass(value))
        and (
            not treat_callable_descriptors_as_non_callable
            or not hasattr(value, '__get__')
        )
    )


class _mocked_property:
    def __init__(self, mock, method_name):
        self.mock = mock
        self.method_name = method_name

    def __get__(self, obj, type):
        # For property/descriptors, `thenCallOriginalImplementation()` must
        # call `original_descriptor.__get__(obj, type)`. Unlike regular method
        # stubs, these `obj/type` binding values are only available during this
        # attribute access path, so we keep them in temporary context.
        with self.mock.property_access_context(self.method_name, obj, type):
            return invocation.RememberedPropertyAccess(
                self.mock, self.method_name)()

    def __set__(self, obj, value):
        # Keep this wrapper a data descriptor so it wins over instance
        # __dict__ during reads.
        return None

    def __delete__(self, obj):
        return None


class Mock:
    def __init__(
        self,
        mocked_obj: object,
        strict: bool = True,
        spec: object | None = None
    ) -> None:
        self.mocked_obj = mocked_obj
        self.strict = strict
        self.spec = spec

        self.invocations: list[invocation.RealInvocation] = []
        self.stubbed_invocations: deque[invocation.StubbedInvocation] = deque()

        self._original_methods: dict[str, object | None] = {}
        self._methods_to_unstub: dict[str, Patch] = {}
        self._signatures_store: dict[str, signature.Signature | None] = {}
        self._property_access_context: \
            list[tuple[str, object | None, object]] = []

        self._continuations: dict[
            invocation.StubbedInvocation,
            invocation.ConfiguredContinuation,
        ] = {}

        self._observers: list = []
        self._methods_marked_as_coroutine: set[str] = set()

    def attach(self, observer) -> None:
        if observer not in self._observers:
            self._observers.append(observer)

    def detach(self, observer) -> None:
        try:
            self._observers.remove(observer)
        except ValueError:
            pass

    def remember(self, invocation: invocation.RealInvocation) -> None:
        self.invocations.append(invocation)
        for observer in self._observers:
            observer.update(invocation)

    def finish_stubbing(
        self, stubbed_invocation: invocation.StubbedInvocation
    ) -> None:
        self.stubbed_invocations.appendleft(stubbed_invocation)

    def clear_invocations(self) -> None:
        self.invocations = []

    def continuation_for(
        self, invoc: invocation.StubbedInvocation
    ) -> invocation.Continuation:
        continuation = self._continuations.get(invoc)
        if continuation is not None:
            return continuation

        for other in self._sameish_invocations(invoc):
            configured_continuation = self._continuations.get(other)
            if isinstance(configured_continuation, invocation.ValueContinuation):
                return configured_continuation

            if isinstance(configured_continuation, invocation.ChainContinuation):
                # We do not keep mixed continuation modes (`Value` and `Chain`)
                # alive at the same time. So the first chain continuation
                # can be returned immediately.
                return configured_continuation

        return invocation.UnconfiguredContinuation()

    def set_continuation(self, continuation: invocation.ConfiguredContinuation) -> None:
        self._continuations[continuation.invocation] = continuation

    def _sameish_invocations(
        self, same: invocation.StubbedInvocation
    ) -> list[invocation.StubbedInvocation]:
        """Find prior stubs that are signature-compatible.

        This is used only for continuation bookkeeping (value-vs-chain mode),
        not for runtime call dispatch. The comparison is structural and avoids
        executing matcher predicates, so `arg_that(...)` and other custom
        matchers cannot crash internal equivalence probing.

        Why this exists: repeated selectors such as

            when(cat).meow().purr()
            when(cat).meow().roll()

        should share the same root continuation for `meow()`.
        """
        sameish: list[invocation.StubbedInvocation] = []
        for invoc in self.stubbed_invocations:
            if invoc is same:
                continue

            if invoc.method_name != same.method_name:
                continue

            if self._invocations_are_sameish(invoc, same):
                sameish.append(invoc)

        return sameish

    def _invocations_are_sameish(
        self,
        left: invocation.StubbedInvocation,
        right: invocation.StubbedInvocation,
    ) -> bool:
        return sameish.invocations_are_sameish(left, right)

    def get_original_method(self, method_name: str) -> object | None:
        return self._original_methods.get(method_name, None)

    def peek_original_method(self, method_name: str) -> object | None:
        try:
            return self._original_methods[method_name]
        except KeyError:
            original_method, _ = self._get_original_method_before_stub(method_name)
            return original_method

    @contextmanager
    def property_access_context(
        self, method_name: str, obj: object | None, type_: object
    ):
        self._property_access_context.append((method_name, obj, type_))
        try:
            yield
        finally:
            self._property_access_context.pop()

    def get_current_property_access(
        self, method_name: str
    ) -> tuple[object | None, object]:
        for accessed_method_name, obj, type_ in reversed(
            self._property_access_context
        ):
            if accessed_method_name == method_name:
                return obj, type_

        raise RuntimeError(
            "Could not resolve property access context for '%s'."
            % method_name
        )

    # STUBBING

    def _get_original_method_before_stub(
        self, method_name: str
    ) -> tuple[object | None, bool]:
        """
        Looks up the original method on the `spec` object and returns it
        together with an indication of whether the method is found
        "directly" on the `spec` object.

        This is used to decide whether the method should be stored as an
        original_method and should therefore be replaced when unstubbing.
        """
        if self.spec is None:
            return None, False

        return utils.get_original_attribute(self.spec, method_name, default=None)

    def replace_method(
        self,
        method_name: str,
        original_method: object | None,
    ) -> Patch:
        discard_first_arg = self._takes_implicit_self_or_cls(original_method)

        def new_mocked_method(*args, **kwargs):
            return remembered_invocation_builder(
                self, method_name, discard_first_arg, *args, **kwargs
            )

        new_mocked_method.__name__ = method_name
        if original_method:
            new_mocked_method.__doc__ = original_method.__doc__
            new_mocked_method.__wrapped__ = original_method  # type: ignore[attr-defined]
            try:
                new_mocked_method.__module__ = original_method.__module__
            except AttributeError:
                pass

        if (
            self.method_expects_awaitable(method_name, original_method)
            and SUPPORTS_MARKCOROUTINEFUNCTION
        ):
            new_mocked_method = inspect.markcoroutinefunction(new_mocked_method)

        if inspect.ismethod(original_method):
            new_mocked_method = utils.newmethod(new_mocked_method, self.mocked_obj)

        if isinstance(original_method, staticmethod):
            new_mocked_method = staticmethod(new_mocked_method)
        elif isinstance(original_method, classmethod):
            new_mocked_method = classmethod(new_mocked_method)  # type: ignore[assignment]  # noqa: E501
        elif (
            inspect.isclass(self.mocked_obj)
            and inspect.isclass(original_method)  # TBC: Inner classes
        ):
            new_mocked_method = staticmethod(new_mocked_method)

        return patcher.patch_attribute(
            self.mocked_obj,
            method_name,
            new_mocked_method,
            allow_unstub_by_replacement=False,
        )

    def stub(self, method_name: str) -> None:
        try:
            self._methods_to_unstub[method_name]
        except KeyError:
            original_method, _ = self._get_original_method_before_stub(method_name)
            self._original_methods[method_name] = original_method
            self._methods_to_unstub[method_name] = self.replace_method(
                method_name,
                original_method,
            )

    def stub_property(self, method_name: str) -> None:
        try:
            self._methods_to_unstub[method_name]
        except KeyError:
            original_method, _ = self._get_original_method_before_stub(method_name)
            self._original_methods[method_name] = original_method
            self._methods_to_unstub[method_name] = patcher.patch_attribute(
                self.mocked_obj,
                method_name,
                _mocked_property(self, method_name),
                allow_unstub_by_replacement=False,
            )


    def forget_stubbed_invocation(
        self, invocation: invocation.StubbedInvocation
    ) -> None:
        assert invocation in self.stubbed_invocations

        self.stubbed_invocations.remove(invocation)
        self._continuations.pop(invocation, None)

        if not any(
            inv.method_name == invocation.method_name
            for inv in self.stubbed_invocations
        ):
            patch = self._methods_to_unstub.pop(invocation.method_name)
            patch.restore_and_unregister()

        if self.stubbed_invocations:
            return

        mock_registry.unstub(self.mocked_obj)

    def unstub_method(self, method_name: str) -> None:
        invocations = [
            invoc
            for invoc in self.stubbed_invocations
            if invoc.method_name == method_name
        ]
        if not invocations:
            return

        for invoc in invocations:
            invoc.forget_self()

        self.invocations = [
            invocation
            for invocation in self.invocations
            if invocation.method_name != method_name
        ]

    def unstub(self) -> None:
        while self._methods_to_unstub:
            _, patch = self._methods_to_unstub.popitem()
            patch.restore_and_unregister()
        self.stubbed_invocations = deque()
        self.invocations = []
        self._methods_marked_as_coroutine = set()
        self._continuations = {}

    # SPECCING

    def mark_as_coroutine(self, method_name: str) -> None:
        self._methods_marked_as_coroutine.add(method_name)

    def is_marked_as_coroutine(self, method_name: str) -> bool:
        return method_name in self._methods_marked_as_coroutine

    def method_expects_awaitable(
        self,
        method_name: str,
        original_method: object | None = None,
    ) -> bool:
        if original_method is None:
            original_method = self.peek_original_method(method_name)

        return (
            _is_coroutine_method(original_method)
            or self.is_marked_as_coroutine(method_name)
            or (
                self.spec is None
                and method_name in _ASYNC_BY_PROTOCOL_METHODS
            )
        )

    def has_method(self, method_name: str) -> bool:
        if self.spec is None:
            return True

        return hasattr(self.spec, method_name)

    def get_signature(self, method_name: str) -> signature.Signature | None:
        if self.spec is None:
            return None

        try:
            return self._signatures_store[method_name]
        except KeyError:
            sig = signature.get_signature(self.spec, method_name)
            self._signatures_store[method_name] = sig
            return sig

    def will_have_self_or_cls(self, method_name: str) -> bool:
        original_method = self.peek_original_method(method_name)
        return self._takes_implicit_self_or_cls(original_method)

    def _takes_implicit_self_or_cls(
        self, original_method: object | None
    ) -> bool:
        # If original_method is None, we *added* it to mocked_obj
        # and thus, it will eat self iff mocked_obj is a class.
        return (
            inspect.ismethod(original_method)
            or (
                inspect.isclass(self.mocked_obj)
                and not isinstance(original_method, staticmethod)
                and not inspect.isclass(original_method)
            )
        )


def _is_coroutine_method(method: object | None) -> bool:
    if isinstance(method, (staticmethod, classmethod)):
        method = method.__func__
    elif inspect.ismethod(method):
        method = method.__func__

    return inspect.iscoroutinefunction(method)


class _OMITTED(object):
    def __repr__(self):
        return 'OMITTED'


OMITTED = _OMITTED()

def mock(config_or_spec=None, spec=None, strict=OMITTED):  # noqa: C901
    """Create 'empty' objects ('Mocks').

    Will create an empty unconfigured object, that you can pass
    around. All interactions (method calls) will be recorded and can be
    verified using :func:`verify` et.al.

    A plain `mock()` will be not `strict`, and thus all methods regardless
    of the arguments will return ``None``.

    .. note:: Technically all attributes will return an internal interface.
        Because of that a simple ``if mock().foo:`` will surprisingly pass.

    If you set strict to ``True``: ``mock(strict=True)`` all unexpected
    interactions will raise an error instead.

    You configure a mock using :func:`when`, :func:`when2` or :func:`expect`.
    You can also very conveniently just pass in a dict here::

        response = mock({'text': 'ok', 'raise_for_status': lambda: None})

    You can also create an empty Mock which is specced against a given
    `spec`: ``mock(requests.Response)``. These mock are by default strict,
    thus they raise if you want to stub a method, the spec does not implement.
    Mockito will also match the function signature.

    You can pre-configure a specced mock as well::

        response = mock({'json': lambda: {'status': 'Ok'}},
                        spec=requests.Response)

    Mocks are by default callable. Configure the callable behavior using
    `when`::

        dummy = mock()
        when(dummy).__call__(1).thenReturn(2)

    All other magic methods must be configured this way or they will raise an
    AttributeError.


    See :func:`verify` to verify your interactions after usage.

    """

    if type(config_or_spec) is dict:
        config, spec = config_or_spec, spec
    else:
        config, spec = {}, spec or config_or_spec

    if strict is OMITTED:
        strict = False if spec is None else True


    class Dummy(_Dummy):
        if spec:
            __class__ = spec  # make isinstance work

        def __getattr__(self, method_name):
            if strict:
                __tracebackhide__ = operator.methodcaller(
                    "errisinstance", AttributeError
                )
                # deepcopy catches a possible AttributeError, fallback
                # to an arbitrary RuntimeError
                error_type = (
                    RuntimeError
                    if method_name == '__deepcopy__'
                    else AttributeError
                )
                raise error_type(
                    "'Dummy' has no attribute %r configured" % method_name)

            if (
                method_name != "__call__"
                and method_name == "__{}__".format(method_name[2:-2])
            ):
                raise AttributeError(method_name)

            # If a descriptor exists on the dummy class, resolve it here so
            # InvocationError from descriptor-backed stubs is not converted
            # into a dynamic fallback attribute.
            if method_name != "__call__":
                try:
                    class_attr = inspect.getattr_static(type(self), method_name)
                except AttributeError:
                    pass
                else:
                    if hasattr(class_attr, "__get__"):
                        try:
                            return class_attr.__get__(self, type(self))
                        except AttributeError as error:
                            if isinstance(error, invocation.InvocationError):
                                raise
                            # Keep dynamic-attribute behavior for descriptors that
                            # deliberately signal missing via AttributeError.

            def ad_hoc_function(*args, **kwargs):
                return remembered_invocation_builder(
                    theMock, method_name, False, *args, **kwargs
                )
            ad_hoc_function.__name__ = method_name
            ad_hoc_function.__self__ = obj  # type: ignore[attr-defined]
            if spec:
                try:
                    original_method = getattr(spec, method_name)
                    ad_hoc_function.__wrapped__ = original_method  # type: ignore[attr-defined]  # noqa: E501
                    ad_hoc_function.__doc__ = original_method.__doc__
                except AttributeError:
                    pass

            return ad_hoc_function

        def __repr__(self):
            name = 'Dummy'
            if spec:
                name += spec.__name__
            return "<%s id=%s>" % (name, id(self))


    # That's a tricky one: The object we will return is an *instance* of our
    # Dummy class, but the mock we register will point and patch the class.
    # T.i. so that magic methods (`__call__` etc.) can be configured.
    obj = Dummy()
    theMock = Mock(Dummy, strict=strict, spec=spec)

    normalized_names = {
        _normalize_config_key(raw_name)[0]
        for raw_name in config
    }

    for raw_name, value in config.items():
        _configure_mock_from_shorthand(
            theMock,
            Dummy,
            obj,
            raw_name,
            value,
            normalized_names,
        )

    mock_registry.register(obj, theMock)
    return obj


def _configure_mock_from_shorthand(
    theMock: Mock,
    Dummy: type,
    obj: object,
    raw_name: str,
    value: object,
    configured_names: set[str],
) -> None:
    method_name, marked_async = _normalize_config_key(raw_name)
    should_be_async = marked_async or method_name in _ASYNC_BY_PROTOCOL_METHODS

    if method_name in {"__enter__", "__aenter__"} and value is Ellipsis:
        _stub_from_shorthand(
            theMock,
            method_name,
            return_value=obj,
            force_async=(method_name == "__aenter__"),
        )

        companion_exit = "__aexit__" if method_name == "__aenter__" else "__exit__"
        if companion_exit not in configured_names:
            _stub_from_shorthand(
                theMock,
                companion_exit,
                return_value=False,
                force_async=(companion_exit == "__aexit__"),
            )
        return

    if method_name == "__iter__":
        iter_answer = _normalize_iter_answer(value)
        _stub_from_shorthand(theMock, method_name, answer=iter_answer)
        return

    if method_name == "__aiter__":
        aiter_answer = _normalize_aiter_answer(value)
        _stub_from_shorthand(theMock, method_name, answer=aiter_answer)
        return

    if inspect.isfunction(value):
        function_answer = _widen_zero_arg_callable(value)
        _stub_from_shorthand(
            theMock,
            method_name,
            answer=function_answer,
            force_async=should_be_async,
        )
        return

    if should_be_async:
        if value is Ellipsis:
            _stub_from_shorthand(theMock, method_name, force_async=True)
            return

        raise TypeError(
            "Async shorthand '%s' expects a function value or Ellipsis. "
            "Use `lambda: value` for fixed async return values."
            % raw_name
        )

    setattr(Dummy, method_name, value)


def _normalize_config_key(raw_name: str) -> tuple[str, bool]:
    if raw_name.startswith(_CONFIG_ASYNC_PREFIX):
        return raw_name[len(_CONFIG_ASYNC_PREFIX):], True
    return raw_name, False


def _stub_from_shorthand(
    theMock: Mock,
    method_name: str,
    *,
    answer: object = OMITTED,
    return_value: object = OMITTED,
    force_async: bool = False,
) -> None:
    if force_async:
        theMock.mark_as_coroutine(method_name)

    stubbed = invocation.StubbedInvocation(theMock, method_name)(Ellipsis)

    if answer is not OMITTED:
        stubbed.thenAnswer(answer)  # type: ignore[arg-type]
    elif return_value is not OMITTED:
        stubbed.thenReturn(return_value)


def _widen_zero_arg_callable(function: object):
    if not inspect.isfunction(function):
        return function

    try:
        params = inspect.signature(function).parameters
    except Exception:
        return function

    if params:
        return function

    def widened(*args, **kwargs):
        return function()

    widened.__name__ = function.__name__
    widened.__doc__ = function.__doc__
    return widened


def _normalize_iter_answer(value) -> Callable[..., Iterator[object]]:
    def answer(*args, **kwargs) -> Iterator[object]:
        result = value(*args, **kwargs) if callable(value) else value
        return iter(cast(Iterable[object], result))

    return answer


def _normalize_aiter_answer(value) -> Callable[..., AsyncIterator[object]]:
    def answer(*args, **kwargs) -> AsyncIterator[object]:
        result = value(*args, **kwargs) if callable(value) else value
        return _normalize_aiter_result(result)

    return answer


def _normalize_aiter_result(value) -> AsyncIterator[object]:
    if hasattr(value, "__anext__"):
        return cast(AsyncIterator[object], value)

    aiter = getattr(value, "__aiter__", None)
    if callable(aiter):
        candidate = aiter()
        if hasattr(candidate, "__anext__"):
            return cast(AsyncIterator[object], candidate)
        raise TypeError(
            "__aiter__() must return an async iterator implementing __anext__"
        )

    iterator = iter(cast(Iterable[object], value))

    async def generator() -> AsyncIterator[object]:
        for item in iterator:
            yield item

    return generator()



# Copyright (c) 2008-2016 Szczepan Faber, Serhiy Oplakanets, Herr Kaste
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import annotations
from collections.abc import Iterable, MutableMapping
import operator

from . import invocation
from . import verification

from .utils import deprecated, get_obj, get_obj_attr_tuple
from .mocking import Chain, Mock
from .mock_registry import mock_registry
from .patching import restore_patch_contextmanager, patcher
from .verification import VerificationError


class ArgumentError(Exception):
    pass


__tracebackhide__ = operator.methodcaller(
    "errisinstance", (ArgumentError, VerificationError)
)


def _invalid_between(between) -> bool:
    if between is not None:
        try:
            if len(between) == 1:
                start, end = between[0], float('inf')
            else:
                start, end = between
            if start < 0 or start > end:
                return True
        except Exception:
            return True
    return False


def _get_wanted_verification(
        times=None, atleast=None, atmost=None, between=None
) -> verification.VerificationMode | None:
    if (times, atleast, atmost, between).count(None) < 3:
        raise ArgumentError(
            "You can set only one of the arguments: 'times', 'atleast', "
            "'atmost' or 'between'.")

    if times is not None:
        if times < 0:
            raise ArgumentError(
                "'times' argument has invalid value.\n"
                f"It should be at least 0.  You wanted to set it to: {times}"
            )
        return verification.Times(times)
    if atleast is not None:
        if atleast < 1:
            raise ArgumentError(
                "'atleast' argument has invalid value.\n"
                f"It should be at least 1.  You wanted to set it to: {atleast}"
            )
        return verification.AtLeast(atleast)
    if atmost is not None:
        if atmost < 1:
            raise ArgumentError(
                "'atmost' argument has invalid value.\n"
                f"It should be at least 1.  You wanted to set it to: {atmost}"
            )
        return verification.AtMost(atmost)
    if between is not None:
        if _invalid_between(between):
            raise ArgumentError(
                "'between' argument has invalid value.\n"
                "It should consist of positive values with second number "
                "greater than first e.g. (1, 4) or (0, 3) or (2, 2), "
                "or a single non-negative number for open-ended range "
                f"e.g. (0,).  You wanted to set it to: {between}"
            )
        return verification.Between(*between)
    return None


def _get_mock(obj: object, strict=True) -> Mock:
    theMock = mock_registry.mock_for(obj)
    if theMock is None:
        theMock = Mock(obj, strict=strict, spec=obj)
        mock_registry.register(obj, theMock)
    return theMock

def _get_mock_or_raise(obj: object) -> Mock:
    theMock = mock_registry.mock_for(obj)
    if theMock is None:
        raise ArgumentError("obj '%s' is not registered" % obj)
    return theMock

def verify(
    obj,
    times=None,
    atleast=None,
    atmost=None,
    between=None,
    inorder=False,
    _factory=None,
):
    """Central interface to verify interactions.

    `verify` uses a fluent interface::

        verify(<obj>, times=2).<method_name>(<args>)

    `args` can be as concrete as necessary. Often a catch-all is enough,
    especially if you're working with strict mocks, bc they throw at call
    time on unwanted, unconfigured arguments::

        from mockito import ANY, ARGS, KWARGS
        when(manager).add_tasks(1, 2, 3)
        ...
        # no need to duplicate the specification; every other argument pattern
        # would have raised anyway.
        verify(manager).add_tasks(1, 2, 3)  # duplicates `when`call
        verify(manager).add_tasks(*ARGS)
        verify(manager).add_tasks(...)       # Py3
        verify(manager).add_tasks(Ellipsis)  # Py2

    """

    if isinstance(obj, str):
        obj = get_obj(obj)

    verification_fn = (
        _get_wanted_verification(
            times=times, atleast=atleast, atmost=atmost, between=between
        ) or verification.Times(1)
    )
    if inorder:
        verification_fn = verification.InOrder(verification_fn)

    theMock = _get_mock_or_raise(obj)

    factory = _factory or invocation.VerifiableInvocation

    class Verify(object):
        def __getattr__(self, method_name):
            return factory(theMock, method_name, verification_fn)

    return Verify()


class _OMITTED(object):
    def __repr__(self):
        return 'OMITTED'


OMITTED = _OMITTED()


def when(obj, strict=True):
    """Central interface to stub functions on a given `obj`

    `obj` should be a module, a class or an instance of a class; it can be
    a Dummy you created with :func:`mock`. ``when`` exposes a fluent interface
    where you configure a stub in three steps::

        when(<obj>).<method_name>(<args>).thenReturn(<value>)

    Compared to simple *patching*, stubbing in mockito requires you to specify
    concrete `args` for which the stub will answer with a concrete `<value>`.
    All invocations that do not match this specific call signature will be
    rejected. They usually throw at call time.

    Stubbing in mockito's sense thus means not only to get rid of unwanted
    side effects, but effectively to turn function calls into constants.

    E.g.::

        # Given ``dog`` is an instance of a ``Dog``
        when(dog).bark('Grrr').thenReturn('Wuff')
        when(dog).bark('Miau').thenRaise(TypeError())

        # With this configuration set up:
        assert dog.bark('Grrr') == 'Wuff'
        dog.bark('Miau')  # will throw TypeError
        dog.bark('Wuff')  # will throw unwanted interaction

    Stubbing can effectively be used as monkeypatching; usage shown with
    the `with` context managing::

        with when(os.path).exists('/foo').thenReturn(True):
            ...

    Most of the time verifying your interactions is not necessary, because
    your code under tests implicitly verifies the return value by evaluating
    it. See :func:`verify` if you need to, see also :func:`expect` to setup
    expected call counts up front.

    If your function is pure side effect and does not return something, you
    can omit the specific answer. The function will then return `None` as by
    default for Python functions::

        when(manager).do_work().thenReturn()
        # However, using `expect` may read better.
        expect(manager).do_work()

    `when` verifies the method name, the expected argument signature, and the
    actual, factual arguments your code under test uses against the original
    object and its function so its easier to spot changing interfaces.

    Sometimes it's tedious to spell out all arguments::

        from mockito import ANY, ARGS, KWARGS
        when(requests).get('http://example.com/', **KWARGS).thenReturn(...)
        when(os.path).exists(ANY)
        when(os.path).exists(ANY(str))

    .. note:: You must :func:`unstub` after stubbing, or use `with`
        statement.

    Set ``strict=False`` to bypass the function signature checks.

    See related :func:`when2` which has a more pythonic interface.

    """

    if isinstance(obj, str):
        obj = get_obj(obj)

    theMock = _get_mock(obj, strict=strict)
    chain = Chain(theMock, {"strict": strict})

    class When(object):
        def __getattr__(self, method_name):
            return chain.new_segment(method_name)

    return When()


def when2(fn, *args, **kwargs):
    """Stub a function call with the given arguments

    Exposes a more pythonic interface than :func:`when`. See :func:`when` for
    more documentation.

    Returns `AnswerSelector` interface which exposes `thenReturn`,
    `thenRaise`, `thenAnswer`, and `thenCallOriginalImplementation` as usual.
    Always `strict`.

    Usage::

        # Given `dog` is an instance of a `Dog`
        when2(dog.bark, 'Miau').thenReturn('Wuff')

    .. note:: You must :func:`unstub` after stubbing, or use `with`
        statement.

    """
    obj, name = get_obj_attr_tuple(fn)
    theMock = _get_mock(obj, strict=True)
    return invocation.StubbedInvocation(theMock, name)(*args, **kwargs)


def patch(fn, attr_or_replacement, replacement=None):
    """Patch/Replace a function.

    This is really like monkeypatching, but *note* that all interactions
    will be recorded and can be verified. That is, using `patch` you stay in
    the domain of mockito.

    Two ways to call this. Either::

        patch(os.path.exists, lambda str: True)  # two arguments
        # OR
        patch(os.path, 'exists', lambda str: True)  # three arguments

    If called with three arguments, the mode is *not* strict to allow *adding*
    methods. If called with two arguments, mode is always `strict`.

    .. note:: You must :func:`unstub` after stubbing, or use `with`
        statement.

    """
    if replacement is None:
        replacement = attr_or_replacement
        return when2(fn, Ellipsis).thenAnswer(replacement)
    else:
        obj, name = fn, attr_or_replacement
        theMock = _get_mock(obj, strict=True)
        return invocation.StubbedInvocation(
            theMock, name, strict=False)(Ellipsis).thenAnswer(replacement)


def patch_attr(obj_or_path, attr_or_replacement, replacement=OMITTED):
    """Patch/replace an attribute with a concrete value.

    Unlike :func:`patch`, this does *not* record interactions and does not
    expose verification. It is intended for simple attribute replacement like
    ``sys.stdout`` or ``sys.argv``.

    Two ways to call this. Either::

        patch_attr('sys.stdout', StringIO())  # two arguments
        # OR
        patch_attr(sys, 'stdout', StringIO())  # three arguments

    ``with`` context management is supported and restores the original value
    on ``__exit__``. ``__enter__`` returns the replacement object.

    .. note:: You must :func:`unstub` after patching, or use `with`
        statement.

    """
    if replacement is OMITTED:
        replacement = attr_or_replacement
        obj, name = get_obj_attr_tuple(obj_or_path)
    else:
        obj, name = obj_or_path, attr_or_replacement

    patch = patcher.patch_attribute(
        obj,
        name,
        replacement,
        allow_unstub_by_replacement=True,
    )
    return restore_patch_contextmanager(patch, replacement)


def patch_dict(mapping_or_path, values=None, *, clear=False, remove=None, **kwargs):
    """Patch/update a dict-like object in place.

    This is a convenience function for test-time dictionary patching,
    especially for mutable global maps like ``os.environ``.

    Usage::

        patch_dict(os.environ, {'USER': 'foo'})
        patch_dict(os.environ, [('USER', 'foo')])
        patch_dict(os.environ, USER='foo')
        patch_dict(os.environ, remove={'USER', 'PATH'})
        patch_dict(os.environ, remove=all)
        patch_dict(os.environ, clear=True)
        patch_dict('os.environ', {'USER': 'foo'})

    ``with`` context management is supported and restores the original mapping
    state on ``__exit__``. ``__enter__`` returns the patched mapping.

    ``values`` can be any value accepted by ``dict(values)``.
    ``kwargs`` are merged into ``values`` and take precedence.

    .. note:: You must :func:`unstub` after patching, or use `with`
        statement.

    """
    mapping = (
        get_obj(mapping_or_path)
        if isinstance(mapping_or_path, str)
        else mapping_or_path
    )

    if not isinstance(mapping, MutableMapping):
        raise TypeError("target must be a mutable mapping")

    if remove is all:
        clear = True
        remove = None

    normalized_remove: tuple[object, ...]
    if remove is None:
        normalized_remove = ()
    elif isinstance(remove, (str, bytes)):
        normalized_remove = (remove,)
    elif not isinstance(remove, Iterable):
        raise TypeError("remove must be iterable, all, or None")
    else:
        normalized_remove = tuple(remove)

    updates = {} if values is None else dict(values)
    updates.update(kwargs)
    patch = patcher.patch_dictionary(
        mapping,
        updates,
        clear=clear,
        remove=normalized_remove,
    )
    return restore_patch_contextmanager(patch, mapping)


def expect(obj, strict=True,
           times=None, atleast=None, atmost=None, between=None):
    """Stub a function call, and set up an expected call count.

    Usage::

        # Given `dog` is an instance of a `Dog`
        expect(dog, times=1).bark('Wuff').thenReturn('Miau')
        dog.bark('Wuff')
        dog.bark('Wuff')  # will throw at call time: too many invocations

        # maybe if you need to ensure that `dog.bark()` was called at all
        verifyExpectedInteractions()

    .. note:: You must :func:`unstub` after stubbing, or use `with`
        statement.

    See :func:`when`, :func:`when2`, :func:`verifyExpectedInteractions`

    """

    if isinstance(obj, str):
        obj = get_obj(obj)

    theMock = _get_mock(obj, strict=strict)

    verification_fn = _get_wanted_verification(
        times=times, atleast=atleast, atmost=atmost, between=between)

    chain = Chain(theMock, {
        "verification": verification_fn,
        "strict": strict,
    })

    class Expect(object):
        def __getattr__(self, method_name):
            return chain.new_segment(method_name)

    return Expect()



def unstub(*objs):
    """Unstubs all stubbed methods, functions, and patched attributes.

    If you don't pass in any argument, *all* registered mocks and
    patched modules, classes etc. will be unstubbed.

    You can also unstub a single method/function target, e.g.::

        unstub(os.path.exists)
        unstub("os.path.exists")
        unstub(cat.meow)

    Or explicitly target one attribute by host and name, e.g.::

        unstub((cat, "meow"))
        unstub(cat, "meow")
        unstub((cat, "meow"), (os.path, "exists"))

    In these cases only the selected attributes are restored, while other stubs
    on the same objects stay active.

    Note that additionally, the underlying registry will be cleaned.
    After an `unstub` you can't :func:`verify` anymore because all
    interactions will be forgotten.
    """

    if not objs:
        mock_registry.unstub_all()
        patcher.unstub_all()
        return

    explicit_attr_targets, generic_targets = _partition_unstub_targets(objs)

    for host, attr_name in explicit_attr_targets:
        _unstub_attr_target(host, attr_name)

    for obj in generic_targets:
        if isinstance(obj, str):
            obj = get_obj(obj)

        if mock_registry.unstub(obj) or patcher.unstub_matching(obj):
            continue

        resolved_target = _resolve_unstub_attr_target(obj)
        if resolved_target is None:
            continue

        host, attr_name = resolved_target
        _unstub_attr_target(host, attr_name)



def _partition_unstub_targets(objs):
    if _is_unstub_attr_pair_arguments(objs):
        host, attr_name = objs
        return [
            _normalize_unstub_attr_target(host, attr_name)
        ], []

    explicit_attr_targets = []
    generic_targets = []

    for obj in objs:
        explicit_attr_target = _coerce_unstub_attr_target_tuple(obj)
        if explicit_attr_target is None:
            generic_targets.append(obj)
            continue

        explicit_attr_targets.append(explicit_attr_target)

    return explicit_attr_targets, generic_targets



def _is_unstub_attr_pair_arguments(objs):
    return (
        len(objs) == 2
        and not isinstance(objs[0], tuple)
        and _looks_like_attr_name(objs[1])
    )



def _coerce_unstub_attr_target_tuple(target):
    if not isinstance(target, tuple) or len(target) != 2:
        return None

    host, attr_name = target
    if not _looks_like_attr_name(attr_name):
        return None

    return _normalize_unstub_attr_target(host, attr_name)



def _normalize_unstub_attr_target(host, attr_name):
    if isinstance(host, str):
        host = get_obj(host)

    return host, attr_name



def _looks_like_attr_name(value):
    return isinstance(value, str) and bool(value) and "." not in value



def _unstub_attr_target(host, attr_name):
    host_mock = mock_registry.mock_for(host)
    if host_mock is not None:
        host_mock.unstub_method(attr_name)

    patcher.unstub_attribute(host, attr_name)



def _resolve_unstub_attr_target(target):
    if not callable(target):
        return None

    host = getattr(target, "__self__", None)
    attr_name = getattr(target, "__name__", None)
    if host is not None and attr_name is not None:
        return host, attr_name

    target_function = _unwrap_unstub_target(target)
    for theMock in mock_registry.get_registered_mocks():
        for method_name, patch in theMock._methods_to_unstub.items():
            replacement = getattr(patch, "replacement", None)
            if _unwrap_unstub_target(replacement) is target_function:
                return theMock.mocked_obj, method_name

    return None



def _unwrap_unstub_target(target):
    if isinstance(target, (staticmethod, classmethod)):
        return target.__func__

    return getattr(target, "__func__", target)


def forget_invocations(*objs):
    """Forget all invocations of given objs.

    If you already *call* mocks during your setup routine, you can now call
    ``forget_invocations`` at the end of your setup, and have a clean
    'recording' for your actual test code. T.i. you don't have
    to count the invocations from your setup code anymore when using
    :func:`verify` afterwards.
    """
    for obj in objs:
        theMock = _get_mock_or_raise(obj)
        theMock.clear_invocations()


def ensureNoUnverifiedInteractions(*objs):
    """Check if any given object has any unverified interaction.

    You can use this after `verify`-ing to ensure no other interactions
    happened.

    Can lead to over-specified tests.
    """
    verifyExpectedInteractions(*objs)

    for obj in objs:
        theMock = _get_mock_or_raise(obj)

        for i in theMock.invocations:
            if not i.verified:
                raise VerificationError("\nUnwanted interaction: %s" % i)


def verifyZeroInteractions(*objs):
    """Verify that no methods have been called on given objs.

    Rarely used because `verify(..., times=0)` is more explicit.  Also:
    strict mocks usually throw early on unexpected, unstubbed invocations.
    For them, there may be no need to verify afterwards.
    `expect(..., times=0)` may also appropriate.

    Partial mocks ('monkeypatched' objects or modules) only look at the
    stubbed invocations as the actual usage gets recorded only for them.
    However, you could use `spy` and inject it.

    """
    for obj in objs:
        theMock = _get_mock_or_raise(obj)

        if len(theMock.invocations) > 0:
            raise VerificationError(
                "\nUnwanted interaction: %s" % theMock.invocations[0])



def verifyExpectedInteractions(*objs):
    """Verifies that expectations set via `expect` are met

    E.g.::

        expect(os.path, times=1).exists(...).thenReturn(True)
        os.path('/foo')
        verifyExpectedInteractions(os.path)  # ok, called once

    If you leave out the argument *all* registered objects will
    be checked.

    .. note:: **DANGERZONE**: If you did not :func:`unstub` correctly,
        it is possible that old registered mocks, from other tests
        leak.

    See related :func:`expect`
    """

    if objs:
        theMocks: Iterable[Mock] = map(_get_mock_or_raise, objs)
    else:
        theMocks = mock_registry.get_registered_mocks()

    for mock in theMocks:
        for i in mock.stubbed_invocations:
            i.verify()


def verifyStubbedInvocationsAreUsed(*objs):
    """Ensure stubs are actually used.

    This functions just ensures that stubbed methods are actually used. Its
    purpose is to detect interface changes after refactorings. It is meant
    to be invoked usually without arguments just before :func:`unstub`.

    """
    if objs:
        theMocks: Iterable[Mock] = map(_get_mock_or_raise, objs)
    else:
        theMocks = mock_registry.get_registered_mocks()


    for mock in theMocks:
        for i in mock.stubbed_invocations:
            i.check_used()


@deprecated(
    "'verifyNoMoreInteractions' is deprecated. "
    "Use 'ensureNoUnverifiedInteractions' instead."
)
def verifyNoMoreInteractions(*objs):
    return ensureNoUnverifiedInteractions(*objs)

verifyNoMoreInteractions.__doc__ = (        # noqa: E305
    ensureNoUnverifiedInteractions.__doc__  # type: ignore[operator]
    + "\n\nDeprecated: Use 'ensureNoUnverifiedInteractions' instead."
)


@deprecated(
    "'verifyNoUnwantedInteractions' is deprecated. "
    "Use 'verifyExpectedInteractions' instead."
)
def verifyNoUnwantedInteractions(*args, **kwargs):
    return verifyExpectedInteractions(*args, **kwargs)

verifyNoUnwantedInteractions.__doc__ = (  # noqa: E305
    verifyExpectedInteractions.__doc__    # type: ignore[operator]
    + "\n\nDeprecated: Use 'verifyExpectedInteractions' instead."
)



# Copyright (c) 2008-2016 Szczepan Faber, Serhiy Oplakanets, Herr Kaste
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import annotations
from abc import ABC
from dataclasses import dataclass
import os
import inspect
import operator
from collections import deque
from typing import TYPE_CHECKING, Union

from . import matchers, sameish, signature
from . import verification as verificationModule
from .mock_registry import mock_registry
from .utils import contains_strict

if TYPE_CHECKING:
    from typing import Any, Callable, NoReturn, Self, TypeVar
    from .mocking import Mock
    T = TypeVar('T')


class InvocationError(AttributeError):
    pass

class AnswerError(AttributeError):
    pass


@dataclass(frozen=True)
class UnconfiguredContinuation:
    pass


@dataclass(frozen=True)
class ValueContinuation:
    invocation: StubbedInvocation


@dataclass(frozen=True)
class ChainContinuation:
    invocation: StubbedInvocation
    chain_mock: Mock


Continuation = Union[UnconfiguredContinuation, ValueContinuation, ChainContinuation]
ConfiguredContinuation = Union[ValueContinuation, ChainContinuation]


__tracebackhide__ = operator.methodcaller(
    "errisinstance",
    (InvocationError, verificationModule.VerificationError)
)


class Invocation(object):
    def __init__(self, mock: Mock, method_name: str) -> None:
        self.mock = mock
        self.method_name = method_name
        self.strict = mock.strict

        self.params: tuple[Any, ...] = ()
        self.named_params: dict[str, Any] = {}

    def _remember_params(self, params: tuple, named_params: dict) -> None:
        self.params = params
        self.named_params = named_params

    def __repr__(self):
        args = [repr(p) if p is not Ellipsis else '...'
                for p in self.params]
        kwargs = ["%s=%r" % (key, val)
                  if key is not matchers.KWARGS_SENTINEL else '**kwargs'
                  for key, val in self.named_params.items()]
        params = ", ".join(args + kwargs)
        return "%s(%s)" % (self.method_name, params)


class RealInvocation(Invocation, ABC):
    def __init__(self, mock: Mock, method_name: str) -> None:
        super(RealInvocation, self).__init__(mock, method_name)
        self.verified = False
        self.verified_inorder = False


class RememberedInvocation(RealInvocation):
    def __init__(
        self, mock: Mock, method_name: str, discard_first_arg: bool = False
    ) -> None:
        super(RememberedInvocation, self).__init__(mock, method_name)
        self.discard_first_arg = discard_first_arg

    def ensure_mocked_object_has_method(self, method_name: str) -> None:
        if not self.mock.has_method(method_name):
            raise InvocationError(
                "You tried to call a method '%s' the object (%s) doesn't "
                "have." % (method_name, self.mock.mocked_obj))

    def ensure_signature_matches(
        self, method_name: str, args: tuple, kwargs: dict
    ) -> None:
        sig = self.mock.get_signature(method_name)
        if not sig:
            return

        signature.match_signature(sig, args, kwargs)

    def __call__(self, *params: Any, **named_params: Any) -> Any | None:
        if self.discard_first_arg:
            params_without_first_arg = params[1:]
        else:
            params_without_first_arg = params
        if self.strict:
            self.ensure_mocked_object_has_method(self.method_name)
            self.ensure_signature_matches(
                self.method_name, params_without_first_arg, named_params)

        self._remember_params(params_without_first_arg, named_params)
        self.mock.remember(self)

        for matching_invocation in self.mock.stubbed_invocations:
            if matching_invocation.matches(self):
                matching_invocation.should_answer(self)
                matching_invocation.capture_arguments(self)
                return matching_invocation.answer_first(
                    *params, **named_params)

        if self.strict:
            stubbed_invocations = [
                invoc
                for invoc in self.mock.stubbed_invocations
                if invoc.method_name == self.method_name
            ]
            raise InvocationError(
                """
Called but not expected:

    %s

Stubbed invocations are:

    %s

"""
                % (
                    self,
                    "\n    ".join(
                        str(invoc) for invoc in reversed(stubbed_invocations)
                    )
                )
            )

        return None


class RememberedPropertyAccess(RememberedInvocation):
    def ensure_mocked_object_has_method(self, method_name):
        return True

    def ensure_signature_matches(self, method_name, args, kwargs):
        return True


class RememberedProxyInvocation(RealInvocation):
    """Remember params and proxy to method of original object.

    Calls method on original object and returns it's return value.
    """
    def __call__(self, *params: Any, **named_params: Any) -> Any:
        self._remember_params(params, named_params)
        self.mock.remember(self)
        obj = self.mock.spec
        try:
            method = getattr(obj, self.method_name)
        except AttributeError:
            raise AttributeError(
                "You tried to call method '%s' which '%s' instance does not "
                "have." % (self.method_name, obj))
        return method(*params, **named_params)



class MatchingInvocation(Invocation, ABC):
    """
    Abstract base class for `RememberedInvocation` and `VerifiableInvocation`.

    Mainly implements `matches` which is used to compare calling signatures
    where placeholders and matchers (like `any()` or `Ellipsis`) are
    interpreted. Here, `self` can contain such special placeholders which then
    consume multiple arguments of the (other) `invocation`.

    """
    @staticmethod
    def compare(p1, p2):
        if p1 is Ellipsis:
            return True

        if isinstance(p1, matchers.Matcher):
            if not p1.matches(p2):
                return False
        elif p1 != p2:
            return False
        return True

    def capture_arguments(self, invocation: RealInvocation) -> None:  # noqa: C901
        """Capture arguments of `invocation` into "capturing" matchers of self.

        This is used in conjunction with "capturing" matchers like
        `ArgumentCaptor`, e.g. `captor`.

        Imagine a `when(obj).method(captor).thenReturn()` configuration.  Now,
        when `obj.method("foo")` is called, "foo" will be passed to
        `captor.capture_value`.

        """
        call_captor = self._get_call_captor()
        if call_captor is not None:
            call_captor.capture_call(invocation.params, invocation.named_params)
            return

        for x, p1 in enumerate(self.params):
            if matchers.is_captor_args_sentinel(p1):
                p1.capture_value(tuple(invocation.params[x:]))
                break

            if isinstance(p1, matchers.Capturing):
                try:
                    p2 = invocation.params[x]
                except IndexError:
                    continue

                p1.capture_value(p2)

        # Explicit keyword matchers (excluding the **kwargs rest placeholder).
        # We use these keys to derive the remaining kwargs for rest-capture.
        fixed_named_keys = {
            key
            for key in self.named_params
            if key is not matchers.KWARGS_SENTINEL
        }
        for key, p1 in self.named_params.items():
            if (
                key is matchers.KWARGS_SENTINEL
                and matchers.is_captor_kwargs_sentinel(p1)
            ):
                p1.capture_value({
                    k: v
                    for k, v in invocation.named_params.items()
                    if k not in fixed_named_keys
                })
                continue

            if isinstance(p1, matchers.Capturing):
                try:
                    p2 = invocation.named_params[key]
                except KeyError:
                    continue

                p1.capture_value(p2)


    def _remember_params(self, params: tuple, named_params: dict) -> None:
        if contains_strict(params, matchers.args):
            raise TypeError('args must be used as *args')

        if (
            contains_strict(params, matchers.kwargs)
            or contains_strict(params, matchers.KWARGS_SENTINEL)
        ):
            raise TypeError('kwargs must be used as **kwargs')

        has_call_captor = (
            any(matchers.is_call_captor(p) for p in params)
            or any(matchers.is_call_captor(v) for v in named_params.values())
        )
        if has_call_captor and (
            len(params) != 1
            or not matchers.is_call_captor(params[0])
            or named_params
        ):
            raise TypeError('call_captor must be used as sole argument')

        def wrap(p):
            if p is any or p is matchers.any_:
                return matchers.any_()
            return p

        self.params = tuple(wrap(p) for p in params)
        self.named_params = {k: wrap(v) for k, v in named_params.items()}

    # Note: matches(a, b) does not imply matches(b, a) because
    # the left side might contain wildcards (like Ellipsis) or matchers.
    # In its current form the right side is a concrete call signature.
    def matches(self, invocation: Invocation) -> bool:  # noqa: C901, E501  (too complex)
        if self.method_name != invocation.method_name:
            return False

        if self._get_call_captor() is not None:
            return True

        for x, p1 in enumerate(self.params):
            if (
                p1 is Ellipsis
                and x == len(self.params) - 1
                and not self.named_params
            ):
                return True

            if p1 is matchers.ARGS_SENTINEL:
                break

            if matchers.is_captor_args_sentinel(p1):
                if not p1.matches(invocation.params[x:]):
                    return False
                break

            try:
                p2 = invocation.params[x]
            except IndexError:
                return False

            if not self.compare(p1, p2):
                return False
        else:
            if len(self.params) != len(invocation.params):
                return False

        # Explicit keyword matchers (excluding the **kwargs rest placeholder).
        # We use these keys to derive the remaining kwargs for rest-matching.
        fixed_named_keys = {
            key
            for key in self.named_params
            if key is not matchers.KWARGS_SENTINEL
        }
        for key, p1 in sorted(
            self.named_params.items(),
            key=lambda k_v: 1 if k_v[0] is matchers.KWARGS_SENTINEL else 0
        ):
            if key is matchers.KWARGS_SENTINEL:
                if matchers.is_captor_kwargs_sentinel(p1):
                    rest_kwargs = {
                        k: v
                        for k, v in invocation.named_params.items()
                        if k not in fixed_named_keys
                    }
                    if not p1.matches(rest_kwargs):
                        return False
                break

            try:
                p2 = invocation.named_params[key]
            except KeyError:
                return False

            if not self.compare(p1, p2):
                return False
        else:
            if len(self.named_params) != len(invocation.named_params):
                return False

        return True

    def _get_call_captor(self):
        if (
            len(self.params) == 1
            and not self.named_params
            and matchers.is_call_captor(self.params[0])
        ):
            return self.params[0]
        return None


class VerifiableInvocation(MatchingInvocation):
    """
    Denotes the function or method signature after `verify` is called.

    I.e.  verify(obj).method(arg1, ...)
                      ^^^^^^^^^^^^^^^^^  VerifiableInvocation denotes this part

    The constructor takes the mock object, which is the registered `Mock` for
    the `obj` in the previous examples, the method name (in the example:
    `method`), and the verification mode (i.e. `verificationModule.Times(1)`).

    In the immediately following `__call__` call, the arguments (`args1, ...`)
    are captured and verified.  For `verify` `__call__` ends the verification
    process, there is no third fluent interface.

    Both calls, `__init__` plus `__call__`, encapsulate a method or function
    call.  But the `__call__` is essentially virtual and can contain
    placeholders and matchers.
    """
    def __init__(
        self,
        mock: Mock,
        method_name: str,
        verification: verificationModule.VerificationMode
    ) -> None:
        super(VerifiableInvocation, self).__init__(mock, method_name)
        self.verification = verification
        self.verification_allows_zero_matches = \
            verification_has_lower_bound_of_zero(self.verification)

    def __call__(self, *params: Any, **named_params: Any) -> None:
        self._remember_params(params, named_params)
        matched_invocations = []
        for invocation in self.mock.invocations:
            if self.matches(invocation):
                self.capture_arguments(invocation)
                matched_invocations.append(invocation)

        self.verification.verify(self, len(matched_invocations))

        # check (real) invocations as verified
        for invocation in matched_invocations:
            invocation.verified = True

        self.maybe_check_stubs_as_used()

    def maybe_check_stubs_as_used(self) -> None:
        """Mark matching stubs as used for explicit zero-lower-bound verifies.

        This is important for follow-up global checks like
        ``verifyStubbedInvocationsAreUsed`` and
        ``ensureNoUnverifiedInteractions``: once a call pattern was explicitly
        verified (including the case where the expected count is zero), those
        checks should not fail only because the corresponding stub has factual
        usage count ``0``.

        In other words, such tests are about verifying *absence* of matching
        invocations, not about requiring that the stub was exercised.
        """
        if self.verification_allows_zero_matches:
            for stub in self.mock.stubbed_invocations:
                # Remember: matches(a, b) does not imply matches(b, a)
                # (see above!), so we check for both
                if stub.matches(self) or self.matches(stub):
                    stub.allow_zero_invocations = True


def verification_has_lower_bound_of_zero(
    verification: verificationModule.VerificationMode | None
) -> bool:
    if (
        isinstance(verification, verificationModule.Times)
        and verification.wanted_count == 0
    ):
        return True

    if isinstance(verification, verificationModule.AtMost):
        return True

    if (
        isinstance(verification, verificationModule.Between)
        and verification.wanted_from == 0
    ):
        return True

    return False


class StubbedInvocation(MatchingInvocation):
    """
    Denotes the function or method signature after `when` or `expect` is
    called, -- the second part of the fluent interface.

    I.e.    when(obj).method(arg1, ...).thenReturn(value1)
          expect(obj).method(arg1, ...).thenReturn(value1)
                      ^^^^^^^^^^^^^^^^^  StubbedInvocation denotes this part

    The constructor takes the mock object, which is the registered `Mock` for
    the `obj` in the previous examples, and the method name (in the example:
    `method`).

    The `verification` argument is only given when `expect` is being used.
    `strict` is used to overrule the `strict` flag of the `mock` object.

    In the immediately following `__call__` call, the arguments (`args1, ...`)
    are captured.  The third part of the fluent interface (`AnswerSelector`)
    is returned.

    Both calls, `__init__` plus `__call__`, encapsulate a method or function
    call.  But the `__call__` is essentially virtual and can contain
    placeholders and matchers.

    The actual stubbing occurs directly in the `__call__` method.  The stubbing
    is delegated to the `mock` object.  In essence, it will likely patch or add
    a replacement callable to `obj`, i.e.
    `setattr(obj, method_name, new_method)`.

    Note about the nomenclature:  In strict OOP languages, we only had
    "methods", but in Python `obj` could be a class, instance, or module --
    generally speaking: a "callable". (I.e. classes are also just callables;
    there is no "new" keyword in Python.)

    """
    def __init__(
        self,
        mock: Mock,
        method_name: str,
        verification: verificationModule.VerificationMode | None = None,
        strict: bool | None = None,
        parent_invocation: StubbedInvocation | None = None,
    ) -> None:
        super(StubbedInvocation, self).__init__(mock, method_name)

        #: Holds the verification set up via `expect`.
        #: The verification will be verified implicitly, while using this stub.
        self.verification = verification

        #: Parent chain invocation for context-managed cleanup propagation.
        #:
        #: When this invocation belongs to a child chain mock (e.g. `.purr()`
        #: after `.meow()`), `forget_self()` may need to recursively forget the
        #: parent chain root invocation if that child chain mock becomes empty.
        self.parent_invocation = parent_invocation

        if strict is not None:
            self.strict = strict

        self.refers_coroutine = mock.method_expects_awaitable(method_name)
        self.discard_first_arg = mock.will_have_self_or_cls(method_name)
        default_answer = (
            return_awaitable(None) if self.refers_coroutine else return_(None)
        )
        self.answers = CompositeAnswer(default_answer=default_answer)

        #: Counts how many times this stub has been 'used'.
        #: A stub gets used, when a real invocation matches its argument
        #: signature, and asks for an answer.
        self.used = 0

        #: Set if `verifyStubbedInvocationsAreUsed` should pass, regardless
        #: of any factual invocation. E.g. set by `expect(..., times=0)`
        self.allow_zero_invocations: bool = \
            verification_has_lower_bound_of_zero(verification)

    def ensure_mocked_object_has_method(self, method_name: str) -> None:
        if not self.mock.has_method(method_name):
            raise InvocationError(
                "You tried to stub a method '%s' the object (%s) doesn't "
                "have." % (method_name, self.mock.mocked_obj))

    def ensure_signature_matches(
        self, method_name: str, args: tuple, kwargs: dict
    ) -> None:
        sig = self.mock.get_signature(method_name)
        if not sig:
            return

        signature.match_signature_allowing_placeholders(sig, args, kwargs)

    def __call__(self, *params: Any, **named_params: Any) -> AnswerSelector:
        if self.strict:
            self.ensure_mocked_object_has_method(self.method_name)
            self.ensure_signature_matches(
                self.method_name, params, named_params)
        self._remember_params(params, named_params)

        self.mock.stub(self.method_name)
        self.mock.finish_stubbing(self)
        return AnswerSelector(self, self.refers_coroutine, self.discard_first_arg)

    def forget_self(self) -> None:
        if self in self.mock.stubbed_invocations:
            self.mock.forget_stubbed_invocation(self)
            self._maybe_forget_parent_chain_invocation()

    def _maybe_forget_parent_chain_invocation(self) -> None:
        if self.parent_invocation is None:
            return

        parent_continuation = self.parent_invocation.get_continuation()
        if (
            isinstance(parent_continuation, ChainContinuation)
            and not parent_continuation.chain_mock.stubbed_invocations
        ):
            self.parent_invocation.forget_self()

    def rollback_if_not_configured_by(
        self,
        continuation: ConfiguredContinuation,
    ) -> None:
        if continuation.invocation is not self:
            self.forget_self()

    def get_continuation(self) -> Continuation:
        return self.mock.continuation_for(self)

    def transition_to_value(self) -> None:
        continuation = self.get_continuation()

        if isinstance(continuation, ChainContinuation):
            # Two examples where this branch is reached:
            # 1) same selector, incompatible mode:
            #       sel = when(cat).meow(); sel.purr(); sel.thenReturn(...)
            #    continuation.invocation is `sel`'s invocation -> no rollback.
            # 2) duplicate selector for an already chained signature:
            #       when(cat).meow().purr(); when(cat).meow().thenReturn(...)
            #    continuation.invocation is the earlier configured invocation ->
            #    rollback provisional duplicate before raising.
            self.rollback_if_not_configured_by(continuation)
            raise InvocationError(
                "'%s' is already configured for chained stubbing."
                % self.method_name
            )

        self.mock.set_continuation(ValueContinuation(self))

    def transition_to_chain(self) -> ChainContinuation:
        continuation = self.get_continuation()

        if isinstance(continuation, ChainContinuation):
            if (
                continuation.invocation is not self
                and sameish.invocations_have_distinct_captors(
                    self,
                    continuation.invocation,
                )
            ):
                self.forget_self()
                raise InvocationError(
                    "'%s' is already configured with a different captor "
                    "instance for the same selector. Reuse the same "
                    "captor() / call_captor() object across chain branches."
                    % self.method_name
                )

            self.rollback_if_not_configured_by(continuation)
            return continuation

        if isinstance(continuation, ValueContinuation):
            self.rollback_if_not_configured_by(continuation)
            raise InvocationError(
                "'%s' is already configured with a direct answer."
                % self.method_name
            )

        chain_root, chain_mock = create_chain_mock()
        answer = (
            return_awaitable(chain_root)
            if self.refers_coroutine
            else return_(chain_root)
        )
        self.add_answer(answer)
        continuation = ChainContinuation(self, chain_mock)
        self.mock.set_continuation(continuation)
        return continuation

    def pop_verification(self) -> verificationModule.VerificationMode | None:
        verification, self.verification = self.verification, None
        return verification

    def add_answer(self, answer: Callable) -> None:
        self.answers.add(answer)

    def answer_first(self, *args: Any, **kwargs: Any) -> Any:
        self.used += 1
        return self.answers.answer(*args, **kwargs)

    def should_answer(self, invocation: RememberedInvocation) -> None:
        verification = self.verification
        if not verification:
            return

        # This check runs before `answer_first`. We add '1' because we want
        # to know if the verification passes if this call gets through.
        actual_count = self.used + 1

        if isinstance(verification, verificationModule.Times):
            if actual_count > verification.wanted_count:
                raise InvocationError(
                    "\nWanted times: %i, actual times: %i"
                    % (verification.wanted_count, actual_count))
        elif isinstance(verification, verificationModule.AtMost):
            if actual_count > verification.wanted_count:
                raise InvocationError(
                    "\nWanted at most: %i, actual times: %i"
                    % (verification.wanted_count, actual_count))
        elif isinstance(verification, verificationModule.Between):
            if actual_count > verification.wanted_to:
                raise InvocationError(
                    "\nWanted between: [%s, %s], actual times: %s"
                    % (verification.wanted_from,
                       verification.wanted_to,
                       actual_count))

        # The way mockito's `verify` works is, that it checks off all 'real',
        # remembered invocations, if they get verified. This is a simple
        # mechanism so that a later `ensureNoUnverifiedInteractions` just has
        # to ensure that all invocations have this flag set to ``True``.
        # For verifications set up via `expect` we want all invocations
        # to get verified 'implicitly', on-the-go, so we set this flag here.
        invocation.verified = True

    def verify(self) -> None:
        if self.verification:
            self.verification.verify(self, self.used)

    def check_used(self) -> None:
        if not self.allow_zero_invocations and self.used < len(self.answers):
            raise verificationModule.VerificationError(
                "\nUnused stub: %s" % self)

class StubbedPropertyAccess(StubbedInvocation):
    def ensure_mocked_object_has_attribute(self, method_name: str) -> None:
        if self.mock.spec is None:
            return

        # Property stubbing is class-only; for instances `stub_property`
        # raises a dedicated guidance error. Skip existence checks here so
        # strict mode does not mask that message for dynamic instance attrs.
        if not inspect.isclass(self.mock.mocked_obj):
            return

        try:
            inspect.getattr_static(self.mock.spec, method_name)
            return
        except AttributeError:
            # Static lookup intentionally avoids descriptor execution, but it
            # does not see dynamic class attrs from metaclass hooks.
            if inspect.isclass(self.mock.spec):
                try:
                    getattr(self.mock.spec, method_name)
                    return
                except AttributeError:
                    pass

        raise InvocationError(
            "You tried to stub an attribute '%s' the object (%s) doesn't "
            "have." % (method_name, self.mock.mocked_obj)
        )

    def __call__(self, *params, **named_params):
        if self.strict:
            self.ensure_mocked_object_has_attribute(self.method_name)
        self._remember_params(params, named_params)

        self.mock.stub_property(self.method_name)
        self.mock.finish_stubbing(self)
        return AnswerSelector(self, self.refers_coroutine, self.discard_first_arg)



def create_chain_mock() -> tuple[object, Mock]:
    from .mocking import mock

    chain_root = mock(strict=True)
    theMock = mock_registry.mock_for(chain_root)
    assert theMock is not None, "Missing chain mock registry entry"
    return chain_root, theMock


def return_(value: T) -> Callable[..., T]:
    def answer(*args, **kwargs) -> T:
        return value
    return answer


def return_awaitable(value: T) -> Callable[..., Any]:
    async def answer(*args, **kwargs) -> T:
        return value
    return answer


def raise_(exception: Exception | type[Exception]) -> Callable[..., NoReturn]:
    def answer(*args, **kwargs) -> NoReturn:
        raise exception
    return answer


def raise_awaitable(
    exception: Exception | type[Exception]
) -> Callable[..., Any]:
    async def answer(*args, **kwargs) -> NoReturn:
        raise exception
    return answer

def discard_self(function: Callable[..., T]) -> Callable[..., T]:
    def function_without_self(*args, **kwargs) -> T:
        args = args[1:]
        return function(*args, **kwargs)
    return function_without_self


def as_awaitable(function: Callable[..., Any]) -> Callable[..., Any]:
    async def function_as_awaitable(*args, **kwargs) -> Any:
        return function(*args, **kwargs)

    return function_as_awaitable


def is_awaitable_when_called(function: Callable[..., Any]) -> bool:
    if inspect.iscoroutinefunction(function):
        return True

    return inspect.iscoroutinefunction(getattr(function, "__call__", None))


class AnswerSelector(object):
    def __init__(
        self,
        invocation: StubbedInvocation,
        expects_awaitable: bool,
        discard_first_arg: bool
    ) -> None:
        self.invocation = invocation
        self.expects_awaitable = expects_awaitable
        self.discard_first_arg = discard_first_arg

    def thenReturn(self, *return_values: Any) -> Self:
        for return_value in return_values or (None,):
            answer = (
                return_awaitable(return_value)
                if self.expects_awaitable
                else return_(return_value)
            )
            self._then(answer)
        return self

    def thenRaise(self, *exceptions: Exception | type[Exception]) -> Self:
        for exception in exceptions or (Exception,):
            answer = (
                raise_awaitable(exception)
                if self.expects_awaitable
                else raise_(exception)
            )
            self._then(answer)
        return self

    def thenAnswer(self, *callables: Callable) -> Self:
        for callable in callables or (return_(None),):
            answer = callable
            if self.discard_first_arg:
                answer = discard_self(answer)
            if self.expects_awaitable and not is_awaitable_when_called(callable):
                answer = as_awaitable(answer)
            self._then(answer)
        return self

    def thenCallOriginalImplementation(self) -> Self:
        answer = self.invocation.mock.get_original_method(
            self.invocation.method_name
        )
        if isinstance(self.invocation, StubbedPropertyAccess):
            if not hasattr(answer, '__get__'):
                self.invocation.forget_self()
                raise AnswerError(
                    "'%s' has no original implementation for '%s'." %
                    (
                        self.invocation.mock.mocked_obj,
                        self.invocation.method_name,
                    )
                )
            self._then(self._property_descriptor_answer(answer))
            return self

        if answer is None:
            self.invocation.forget_self()
            raise AnswerError(
                "'%s' has no original implementation for '%s'." %
                (self.invocation.mock.mocked_obj, self.invocation.method_name)
            )

        if (
            # A classmethod is not callable
            # and a staticmethod is not callable in old version of python,
            # so we get the underlying function.
            isinstance(answer, classmethod) or isinstance(answer, staticmethod)
            # If the method is bound, we unbind it.
            or inspect.ismethod(answer)
        ):
            answer = answer.__func__

        # `answer` is runtime-validated by stubbing setup and optional
        # unwrapping above, but mypy still sees `object` here.
        self._then(answer)  # type: ignore[arg-type]
        return self

    def __enter__(self) -> None:
        pass

    def __exit__(self, *exc_info) -> None:
        try:
            self.invocation.verify()
            if os.environ.get("MOCKITO_CONTEXT_MANAGERS_CHECK_USAGE", "1") == "1":
                self.invocation.check_used()
        finally:
            self.invocation.forget_self()

    def _property_descriptor_answer(self, descriptor: Any) -> Callable:
        def answer(*args: Any, **kwargs: Any) -> Any:
            obj, type_ = self.invocation.mock.get_current_property_access(
                self.invocation.method_name
            )
            # Guarded by `hasattr(descriptor, '__get__')` in caller.
            return descriptor.__get__(obj, type_)

        return answer

    def _then(self, answer: Callable) -> None:
        self.invocation.transition_to_value()
        self.invocation.add_answer(answer)



class CompositeAnswer(object):
    def __init__(self, default_answer: Callable = return_(None)) -> None:
        #: Container for answers, which are just ordinary callables
        self.answers: deque[Callable] = deque()
        self.default_answer = default_answer

        #: Counter for the maximum answers we ever had
        self.answer_count = 0

    def __len__(self) -> int:
        # The minimum is '1' bc we always have a default answer of 'None'
        return max(1, self.answer_count)

    def add(self, answer: Callable) -> None:
        self.answer_count += 1
        self.answers.append(answer)

    def answer(self, *args: Any, **kwargs: Any) -> Any:
        if len(self.answers) == 0:
            return self.default_answer(*args, **kwargs)

        if len(self.answers) == 1:
            a = self.answers[0]
        else:
            a = self.answers.popleft()

        return a(*args, **kwargs)




# Copyright (c) 2008-2016 Szczepan Faber, Serhiy Oplakanets, Herr Kaste
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import annotations
import weakref
from typing import TYPE_CHECKING, Callable, Generic, TypeVar

if TYPE_CHECKING:
    from .mocking import Mock


RegisterObserver = Callable[[object, "Mock"], None]
K = TypeVar("K")
V = TypeVar("V")
T = TypeVar("T")


class MockRegistry:
    """Registry for mocks

    Registers mock()s, ensures that we only have one mock() per mocked_obj, and
    iterates over them to unstub each stubbed method.
    """

    def __init__(self) -> None:
        self.mocks: IdentityMap[object, Mock] = IdentityMap()
        self._register_observers: list[weakref.WeakMethod] = []

    def register(self, obj: object, mock: Mock) -> None:
        self.mocks[obj] = mock

        alive_observers: list[weakref.WeakMethod] = []
        for observer_ref in self._register_observers:
            observer = observer_ref()
            if observer is None:
                continue

            observer(obj, mock)
            alive_observers.append(observer_ref)

        self._register_observers = alive_observers

    def add_register_observer(self, observer: RegisterObserver) -> None:
        self._prune_dead_register_observers()
        for observer_ref in self._register_observers:
            callback = observer_ref()
            if callback is not None and callback == observer:
                return

        self._register_observers.append(weakref.WeakMethod(observer))

    def remove_register_observer(self, observer: RegisterObserver) -> None:
        self._prune_dead_register_observers()

        for i, observer_ref in enumerate(self._register_observers):
            callback = observer_ref()
            if callback is not None and callback == observer:
                del self._register_observers[i]
                break

    def _prune_dead_register_observers(self) -> None:
        self._register_observers = [
            observer_ref
            for observer_ref in self._register_observers
            if observer_ref() is not None
        ]

    def mock_for(self, obj: object) -> Mock | None:
        return self.mocks.get(obj, None)

    def obj_for(self, mock: Mock) -> object | None:
        return self.mocks.lookup(mock)

    def unstub(self, obj: object) -> bool:
        try:
            mock = self.mocks.pop(obj)
        except KeyError:
            return False
        else:
            mock.unstub()
            return True

    def unstub_mock(self, mock: Mock) -> None:
        self.mocks.pop_value(mock)
        mock.unstub()

    def unstub_all(self) -> None:
        for mock in self.get_registered_mocks():
            mock.unstub()
        self.mocks.clear()

    def get_registered_mocks(self) -> list[Mock]:
        return self.mocks.values()


# We have this dict like because we want non-hashable items in our registry.
class IdentityMap(Generic[K, V]):
    def __init__(self) -> None:
        self._store: list[tuple[K, V]] = []

    def __setitem__(self, key: K, value: V) -> None:
        for i, (k, _) in enumerate(self._store):
            if k is key:
                self._store[i] = (key, value)
                return
        self._store.append((key, value))

    def remove(self, key: K) -> None:
        for i, (k, _) in enumerate(self._store):
            if k is key:
                del self._store[i]
                return

    def pop(self, key: K) -> V:
        for i, (k, value) in enumerate(self._store):
            if k is key:
                del self._store[i]
                return value
        raise KeyError()

    def pop_value(self, value: V) -> V:
        for i, (key, val) in enumerate(self._store):
            if val is value:
                del self._store[i]
                return val
        raise KeyError()

    def get(self, key: K, default: T | None = None) -> V | T | None:
        for k, value in self._store:
            if k is key:
                return value
        return default

    def lookup(self, value: V, default: T | None = None) -> K | T | None:
        for key, v in self._store:
            if v is value:
                return key
        return default

    def values(self) -> list[V]:
        return [v for k, v in self._store]

    def clear(self) -> None:
        self._store[:] = []


mock_registry = MockRegistry()
